#include <stdio.h>
#include <string.h>
#include <stdlib.h>

int main () {

	char line[1024];

	int outNbr;

	while (fgets(line, 1024,stdin)) {

		line[strlen(line)-1] = 0;

		if (strncmp(line,"Source Output",13) == 0 ) {
			outNbr = atoi(&line[15]);
			continue;
			}

		if (strstr(line,"module-loopback.c")) {
			while (fgets(line, 1024,stdin)) {

				line[strlen(line)-1] = 0;

				if (!strlen(line)) break;

				if (strstr(line,"media.name")) {

					if (strstr(line,"Null"))
						printf("%d Null\n",outNbr);
					else
						printf("%d Loopback\n",outNbr);
					continue;
					}
				}

			}
		}

	}

			
